use utf8;
use strict;
use warnings FATAL => 'all';

package Xyzzy::Document;

use XML::LibXML;
use XML::LibXML::XPathContext;

use base qw(XML::LibXML::Document);

sub new {
	my $class = shift;
	my $self = $class->SUPER::new(@_);
	bless $self, $class;
}

sub old {
	my $class = shift;
	my $self = shift;
	bless $self, $class;
}

sub xhtml {
	my $doc = new Xyzzy::Document('1.0', 'UTF-8');

	my $dtd = $doc->createInternalSubset('html',
		'-//W3C//DTD XHTML 1.1//EN',
		'http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd');

	my $html = $doc->createElementNS('http://www.w3.org/1999/xhtml', 'html');
	$html->setAttribute('xml:lang', 'en');
	$doc->setDocumentElement($html);

	my $head = $doc->createElement('head');
	$head->setAttribute('id', 'head');
	$html->appendChild($head);

	my $body = $doc->createElement('body');
	$body->setAttribute('id', 'body');
	$html->appendChild($body);

	return $doc
}

sub xhtml_forcepretty {
	0
}

our %structuralnodes;
@structuralnodes{qw(html head body div table tr ul ol dl form select)} = ();

sub killnodes {
	foreach my $node (@_) {
		my $type = ref $node;
		if($type eq 'XML::LibXML::Text') {
			my $parent = $node->parentNode;
			if(defined $parent) {
				if(exists $structuralnodes{$parent->localName}) {
					my $text = $node->data;
					$parent->removeChild($node)
						if $text =~ /^\s*$/
				}
			}
#		} elsif($type eq 'XML::LibXML::Comment') {
#			$node->parentNode->removeChild($node)
		} else {
			killnodes($node->childNodes)
		}
	}
}

sub xhtml_cleanup {
	my $self = shift;

#	foreach($self->documentElement->findnodes("/*[local-name()='html']|/*[local-name()='html']/*|/*[local-name()='html']/head//*")) {
	foreach($self->documentElement->findnodes('/*|/*/*|/*/head//*')) {
		$_->removeAttribute('id')
			if $_->hasAttributes
	}
}

sub _gethash() {
	return $_[0]->{$_[1]}
}

sub getElementById {
	my ($self, $id) = @_;

#	our $expr ||= new XML::LibXML::XPathExpression('//*[@id=$id]');
	my $expr = '//*[@id=$id]';

	our $ctx ||= new XML::LibXML::XPathContext;

	our $prev;
	$ctx->registerVarLookupFunc(\&_gethash, {id => $id})
		unless defined $prev && $prev eq $id;

	$prev = $id;

	my $nodes = $ctx->find($expr, $self->documentElement);

	die "Document ID '$id' not unique"
		if @$nodes > 1;

	return $nodes->[0]
}

sub xmlParser {
	our $xmlparser ||= new XML::LibXML;
	$xmlparser->validation(0);
	$xmlparser->load_ext_dtd(0);
	$xmlparser->expand_entities(0);
	$xmlparser->pedantic_parser(1);
	return $xmlparser;
}

sub getChunks {
	my ($self, $file, $cachethis) = @_;
	our %templatecache;

	if(exists $templatecache{$file}) {
		return $templatecache{$file};
	} else {
#		print STDERR '['.localtime, '] [notice] ', ref $self, " parsing $file\n";
		my $parser = $self->xmlParser;
		my $newdoc = $parser->parse_file($file);

		my @nodes = grep { $_->hasAttributes } $newdoc->documentElement->childNodes;
		killnodes(@nodes)
			if $cachethis || $self->xhtml_forcepretty;
		$templatecache{$file} = \@nodes
			if $cachethis;
		return \@nodes;
	}
}

sub getFragment {
	my ($self, $file, $id) = @_;
	our %fragmentcache;

	my $fragment;
	if(exists $fragmentcache{$file}) {
		$fragment = $fragmentcache{$file}{$id};
	} else {
		my $nodes = $self->getChunks($file, 1);
		my %fragments = map {
				my $i = $_->getAttribute('id');
				$i ? ($i, $_) : ()
			} @$nodes;
		$fragmentcache{$file} = \%fragments;
		$fragment = $fragments{$id};
	}
	die "Fragment '$id' not in '$file'\n"
		unless defined $fragment;

	$fragment = $fragment->cloneNode(1);
	$self->adoptNode($fragment);

#	our $expr ||= new XML::LibXML::XPathExpression('.//*[@id]');
	my $expr = './/*[@id]';
	my %ids;
	foreach my $node ($fragment->findnodes($expr)) {
		$ids{$node->getAttribute('id')} = $node;
		$node->removeAttribute('id');
	}
	return ($fragment, \%ids);
}

sub addChunks {
	my $self = shift;
	my $cachethis = shift;

	foreach my $file (@_) {
		my $nodes = $self->getChunks($file, $cachethis);
		
		foreach my $node (@$nodes) {
			my $id = $node->getAttribute('id');
			my $victim = $self->getElementById($id);
			next unless defined $victim;
			my $newnode = $node->cloneNode(1);
			$self->adoptNode($newnode);
			$victim->replaceNode($newnode);
		}
	}
}

1;
